/*
 * $XConsortium: miwidelineZ.c,v 1.3 93/10/29 17:12:45 takamori Exp $
 *
 * Copyright 1988 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * Author:  Keith Packard, MIT X Consortium
 */

/*
 * Mostly integer wideline code.  Uses a technique similar to
 * bresenham zero-width lines, except walks an X edge
 */

#include <stdio.h>
#include <math.h>
#include "X.h"
#include "windowstr.h"
#include "gcstruct.h"
#include "miscstruct.h"
#include "miwidelineZ.h"
#include "ddpex3.h"

#if (defined(SVR4) || defined(SYSV) && defined(SYSV386)) && __STDC__
extern double hypot(double, double);
#endif

#ifdef ICEILTEMPDECL
ICEILTEMPDECL
#endif

/*
 * spans-based polygon filler
 */

void
miZFillPolyHelper (zbs_ptr, pDrawable, pGC, pixel, spanData, y, overall_height,
		   left, right, left_count, right_count)
    ddZBufferStruct *zbs_ptr;
    DrawablePtr	pDrawable;
    GCPtr	pGC;
    unsigned long   pixel;
    ZSpanDataPtr	spanData;
    int		y;			/* start y coordinate */
    int		overall_height;		/* height of entire segment */
    ZPolyEdgePtr	left, right;
    int		left_count, right_count;
{
    register int left_x, left_e, left_z, left_ez;
    int	left_stepx, left_stepz;
    int	left_signdx, left_signdz;
    int	left_dy, left_dx, left_dz;

    register int right_x, right_e, right_z, right_ez;
    int	right_stepx, right_stepz;
    int	right_signdx, right_signdz;
    int	right_dy, right_dx, right_dz;

    int	height;
    int	left_height, right_height;

    register ddCoord3DS *ppt;
    ddCoord3DS *pptInit;
    register int *pwidth, *pdepth;
    int *pwidthInit, *pdepthInit;
    unsigned long oldPixel;
    int		xorg;
    ZSpans	spanRec;

    left_height = 0;
    right_height = 0;
    
    if (!spanData)
    {
   	pptInit = (ddCoord3DS *) ALLOCATE_LOCAL (overall_height * sizeof(*ppt));
    	if (!pptInit)
	    return;
    	pwidthInit = (int *) ALLOCATE_LOCAL (overall_height * sizeof(*pwidth));
    	if (!pwidthInit)
    	{
	    DEALLOCATE_LOCAL (pptInit);
	    return;
    	}
    	pdepthInit = (int *) ALLOCATE_LOCAL (overall_height * sizeof(*pdepth));
    	if (!pdepthInit)
    	{
	    DEALLOCATE_LOCAL (pptInit);
	    DEALLOCATE_LOCAL (pwidthInit);
	    return;
    	}
	ppt = pptInit;
	pwidth = pwidthInit;
	pdepth = pdepthInit;
    	oldPixel = pGC->fgPixel;
    	if (pixel != oldPixel)
    	{
    	    DoChangeGC (pGC, GCForeground, (XID *)&pixel, FALSE);
    	    ValidateGC (pDrawable, pGC);
    	}
    }
    else
    {
	spanRec.points = (ddCoord3DS *) xalloc (overall_height * sizeof (*ppt));
	if (!spanRec.points)
	    return;
	spanRec.widths = (int *) xalloc (overall_height * sizeof (int));
	if (!spanRec.widths)
	{
	    xfree (spanRec.points);
	    return;
	}
	spanRec.depths = (int *) xalloc (overall_height * sizeof (int));
	if (!spanRec.depths)
	{
	    xfree (spanRec.points);
	    xfree (spanRec.widths);
	    return;
	}
	ppt = spanRec.points;
	pwidth = spanRec.widths;
	pdepth = spanRec.depths;
    }

    xorg = 0;
    if (pGC->miTranslate)
    {
	y += pDrawable->y;
	xorg = pDrawable->x;
    }
    while ((left_count || left_height) &&
	   (right_count || right_height))
    {
	MIPOLYRELOADLEFTZ
	MIPOLYRELOADRIGHTZ

	height = left_height;
	if (height > right_height)
	    height = right_height;

	left_height -= height;
	right_height -= height;

	while (--height >= 0)
	{
	    if (right_x >= left_x)
	    {
	        ppt->z = left_z;
		ppt->y = y;
		ppt->x = left_x + xorg;
		ppt++;
		*pwidth++ = right_x - left_x;
		*pdepth++ = right_z - left_z;
	    }
    	    y++;
    	
	    MIPOLYSTEPLEFTZ

	    MIPOLYSTEPRIGHTZ
	}
    }
    if (!spanData)
    {
        ZBuffer_FillSpans(zbs_ptr, pDrawable, pGC,
			  ppt-pptInit, pptInit, pwidthInit, pdepthInit);
    	DEALLOCATE_LOCAL (pdepthInit);
    	DEALLOCATE_LOCAL (pwidthInit);
    	DEALLOCATE_LOCAL (pptInit);
    	if (pixel != oldPixel)
    	{
	    DoChangeGC (pGC, GCForeground, (XID *)&oldPixel, FALSE);
	    ValidateGC (pDrawable, pGC);
    	}
    }
    else
    {
	spanRec.count = ppt - spanRec.points;
	AppendSpanGroupZ (pGC, pixel, &spanRec, spanData);
    }
}

static void
miZFillRectPolyHelper (zbs_ptr, pDrawable, pGC, pixel, spanData, x, y, z, w, h, d)
    ddZBufferStruct *zbs_ptr;
    DrawablePtr	pDrawable;
    GCPtr	pGC;
    unsigned long   pixel;
    ZSpanDataPtr    spanData;
    int		x, y, z, w, h, d;
{
    register ddCoord3DS *ppt;
    ddCoord3DS *pptInit;
    register int *pwidth, *pdepth;
    int *pwidthInit, *pdepthInit;
    unsigned long oldPixel;
    ZSpans	spanRec;
    xRectangle  rect;
    int		xorg;

    if (!spanData)
    {
	  /* In the non-ZBuffering version, this section involves a call
	     to the GC op PolyFillRect.  I have replaced it with a ZBuffer-ed
	     spanfill procedure call */
        pptInit = ppt = (ddCoord3DS *) ALLOCATE_LOCAL (h*sizeof (ddCoord3DS));
        pwidthInit = pwidth = (int *) ALLOCATE_LOCAL (h*sizeof (int));
        pdepthInit = pdepth = (int *) ALLOCATE_LOCAL (h * sizeof (int));
	if (!ppt || !pwidth || !pdepth)
	{
	    if (ppt) DEALLOCATE_LOCAL(pptInit);
	    if (pwidth) DEALLOCATE_LOCAL(pwidthInit);
	    if (pdepth) DEALLOCATE_LOCAL(pdepthInit);
	    return;
	}
    	if (pGC->miTranslate)
    	{
	    y += pDrawable->y;
	    x += pDrawable->x;
    	}
	while (h--)
	{
	    ppt->x = x;
	    ppt->y = y;
	    ppt->z = z;
	    ppt++;
	    *pwidth++=w;
	    *pdepth++=d;
	    y++;
	}
	ZBuffer_FillSpans(zbs_ptr, pDrawable, pGC, 
			  ppt-pptInit, pptInit, pwidthInit, pdepthInit);
	DEALLOCATE_LOCAL(pptInit);
	DEALLOCATE_LOCAL(pwidthInit);
	DEALLOCATE_LOCAL(pdepthInit);
    }
    else
    {
	ZSpanGroup   *group;
	spanRec.points = (ddCoord3DS *) xalloc (h * sizeof (*ppt));
	if (!spanRec.points)
	    return;
	spanRec.widths = (int *) xalloc (h * sizeof (int));
	if (!spanRec.widths)
	{
	    xfree (spanRec.points);
	    return;
	}
	spanRec.depths = (int *) xalloc (h * sizeof (int));
	if (!spanRec.depths)
	{
	    xfree (spanRec.points);
	    xfree (spanRec.widths);
	    return;
	}
	ppt = spanRec.points;
	pwidth = spanRec.widths;
	pdepth = spanRec.depths;

    	if (pGC->miTranslate)
    	{
	    y += pDrawable->y;
	    x += pDrawable->x;
    	}
	while (h--)
	{
	    ppt->x = x;
	    ppt->y = y;
	    ppt->z = z;
	    ppt++;
	    *pwidth++ = w;
	    *pdepth++ = d;
	    y++;
	}
	spanRec.count = ppt - spanRec.points;
	AppendSpanGroupZ (pGC, pixel, &spanRec, spanData);
    }
}

static int
miZPolyBuildEdge (x0, y0, z0, k, kz, dx, dy, dz, xi, yi, zi, left, edge)
    double	x0, y0, z0;
    double	k;  /* x0 * dy - y0 * dx */
    double	kz; /* z0 * dy - y0 * dz */
    register int dx, dy, dz;
    int		xi, yi, zi;
    int		left;
    register ZPolyEdgePtr edge;
{
    int	    x, y, z, e, ez;
    int	    xady, zady;

    if (dy < 0)
    {
        dz = -dz;
	dy = -dy;
	dx = -dx;
	k = -k;
	kz = -kz;
    }

#ifdef NOTDEF
    {
	double	realk, kerror, realkz, kzerror;
    	realk = x0 * dy - y0 * dx;
    	realkz = z0 * dy - y0 * dz;
    	kerror = fabs (realk - k);
    	if (kerror > .1)
	    printf ("realk: %g k: %g\n", realk, k);
    	kzerror = fabs (realkz - kz);
    	if (kzerror > .1)
	    printf ("realkz: %g kz: %g\n", realkz, kz);
    }
#endif
    y = ICEIL (y0);
    xady = ICEIL (k) + y * dx;
    zady = ICEIL (kz) + y * dz;

    if (xady <= 0)
	x = - (-xady / dy) - 1;
    else
	x = (xady - 1) / dy;
    if (zady <= 0)
	z = - (-zady / dy) - 1;
    else
	z = (zady - 1) / dy;

    e = xady - x * dy;
    ez = zady - z * dy;

    if (dx >= 0)
    {
	edge->signdx = 1;
	edge->stepx = dx / dy;
	edge->dx = dx % dy;
    }
    else
    {
	edge->signdx = -1;
	edge->stepx = - (-dx / dy);
	edge->dx = -dx % dy;
	e = dy - e + 1;
    }
    if (dz >= 0)
    {
	edge->signdz = 1;
	edge->stepz = dz / dy;
	edge->dz = dz % dy;
    }
    else
    {
	edge->signdz = -1;
	edge->stepz = - (-dz / dy);
	edge->dz = -dz % dy;
	ez = dy - ez + 1;
    }
    edge->dy = dy;
    edge->x = x + left + xi;
    edge->e = e - dy;	/* bias to compare against 0 instead of dy */
    edge->z = z + zi;
    edge->ez = ez - dy;	/* bias to compare against 0 instead of dy */
    return y + yi;
}

#define StepAround(v, incr, max) (((v) + (incr) < 0) ? (max - 1) : ((v) + (incr) == max) ? 0 : ((v) + (incr)))

static int
miZPolyBuildPoly (vertices, slopes, count, xi, yi, zi, left, right, pnleft, pnright, h)
    register ZPolyVertexPtr vertices;
    register ZPolySlopePtr  slopes;
    int		    count;
    int		    xi, yi, zi;
    ZPolyEdgePtr	    left, right;
    int		    *pnleft, *pnright;
    int		    *h;
{
    int	    top, bottom;
    double  miny, maxy;
    register int i;
    int	    j;
    int	    clockwise;
    int	    slopeoff;
    register int s;
    register int nright, nleft;
    int	    y, lasty, bottomy, topy;

    /* find the top of the polygon */
    maxy = miny = vertices[0].y;
    bottom = top = 0;
    for (i = 1; i < count; i++)
    {
	if (vertices[i].y < miny)
	{
	    top = i;
	    miny = vertices[i].y;
	}
	if (vertices[i].y >= maxy)
	{
	    bottom = i;
	    maxy = vertices[i].y;
	}
    }
    clockwise = 1;
    slopeoff = 0;

    i = top;
    j = StepAround (top, -1, count);

    if (slopes[j].dy * slopes[i].dx > slopes[i].dy * slopes[j].dx)
    {
	clockwise = -1;
	slopeoff = -1;
    }

    bottomy = ICEIL (maxy) + yi;

    nright = 0;

    s = StepAround (top, slopeoff, count);
    i = top;
    while (i != bottom)
    {
	if (slopes[s].dy != 0)
	{
	    y = miZPolyBuildEdge (vertices[i].x, vertices[i].y, vertices[i].z,
			slopes[s].k, slopes[s].kz,
			slopes[s].dx, slopes[s].dy, slopes[s].dz,
			xi, yi, zi, 0,
			&right[nright]);
	    if (nright != 0)
	    	right[nright-1].height = y - lasty;
	    else
	    	topy = y;
	    nright++;
	    lasty = y;
	}

	i = StepAround (i, clockwise, count);
	s = StepAround (s, clockwise, count);
    }
    if (nright != 0)
	right[nright-1].height = bottomy - lasty;

    if (slopeoff == 0)
	slopeoff = -1;

    else
	slopeoff = 0;

    nleft = 0;
    s = StepAround (top, slopeoff, count);
    i = top;
    while (i != bottom)
    {
	if (slopes[s].dy != 0)
	{
	    y = miZPolyBuildEdge (vertices[i].x, vertices[i].y, vertices[i].z,
				  slopes[s].k, slopes[s].kz,
				  slopes[s].dx,  slopes[s].dy, slopes[s].dz, 
				  xi, yi, zi, 1,
				  &left[nleft]);
    
	    if (nleft != 0)
	    	left[nleft-1].height = y - lasty;
	    nleft++;
	    lasty = y;
	}
	i = StepAround (i, -clockwise, count);
	s = StepAround (s, -clockwise, count);
    }
    if (nleft != 0)
	left[nleft-1].height = bottomy - lasty;
    *pnleft = nleft;
    *pnright = nright;
    *h = bottomy - topy;
    return topy;
}

static void
miZLineOnePoint (zbs_ptr, pDrawable, pGC, pixel, spanData, x, y, z)
    ddZBufferStruct *zbs_ptr;
    DrawablePtr	    pDrawable;
    GCPtr	    pGC;
    unsigned long   pixel;
    ZSpanDataPtr    spanData;
    int		    x, y, z;
{
    ddCoord2DS pt;
    int	    wid, dep;
    unsigned long	oldPixel;

    MILINESETPIXEL (pDrawable, pGC, pixel, oldPixel);
    if (pGC->fillStyle == FillSolid)
    {
	pt.x = x;
	pt.y = y;
	dep = z;
	(*pGC->ops->PolyPoint) (pDrawable, pGC, CoordModeOrigin, 1, &pt);
    }
    else
    {
	wid = 1;
	if (pGC->miTranslate) 
	{
	    x += pDrawable->x;
	    y += pDrawable->y;
	}
	pt.x = x;
	pt.y = y;
	dep = z;
	ZBuffer_FillSpans(zbs_ptr, pDrawable, pGC,
			  1, &pt, &wid, &dep);
    }
    MILINERESETPIXEL (pDrawable, pGC, pixel, oldPixel);
}

static void
miZLineJoin (zbs_ptr, pDrawable, pGC, pixel, spanData, pLeft, pRight)
    ddZBufferStruct *zbs_ptr;
    DrawablePtr	    pDrawable;
    GCPtr	    pGC;
    unsigned long   pixel;
    ZSpanDataPtr    spanData;
    register ZLineFacePtr pLeft, pRight;
{
    double	    mx, my;
    int		    denom;
    ZPolyVertexRec   vertices[4];
    ZPolySlopeRec    slopes[4];
    int		    edgecount;
    ZPolyEdgeRec	    left[4], right[4];
    int		    nleft, nright;
    int		    y, height;
    int		    swapslopes;
    int		    joinStyle = pGC->joinStyle;
    int		    lw = pGC->lineWidth;
    double          scale, dx, dy, dz, adx, ady, adz;           /* scale to int, delta and absolute delta values across the left and right */

    if (lw == 1 && !spanData) {
	/* Lines going in the same direction have no join */
	if (pLeft->dx >= 0 == pRight->dx <= 0)
	    return;
	denom = - pLeft->dx * pRight->dy + pRight->dx * pLeft->dy;
	if (denom == 0)
	    return;	/* no join to draw */
	miZLineOnePoint (zbs_ptr, pDrawable, pGC, pixel, spanData, pLeft->x, pLeft->y, pLeft->z);
	return;
    }
    denom = - pLeft->dx * pRight->dy + pRight->dx * pLeft->dy;
    if (denom == 0)
        return;	/* no join to draw */

    swapslopes = 0;
    if (denom > 0)
    {
	pLeft->xa = -pLeft->xa;
	pLeft->ya = -pLeft->ya;
	pLeft->za = -pLeft->za;
	pLeft->dx = -pLeft->dx;
	pLeft->dy = -pLeft->dy;
	pLeft->dz = -pLeft->dz;
    }
    else
    {
	swapslopes = 1;
	pRight->xa = -pRight->xa;
	pRight->ya = -pRight->ya;
	pRight->za = -pRight->za;
	pRight->dx = -pRight->dx;
	pRight->dy = -pRight->dy;
	pRight->dz = -pRight->dz;
    }

    vertices[0].x = pRight->xa;
    vertices[0].y = pRight->ya;
    vertices[0].z = pRight->za;
    slopes[0].dx = -pRight->dy;
    slopes[0].dy =  pRight->dx;
    slopes[0].dz = -pRight->dy;
    slopes[0].k = 0;
    slopes[0].kz = 0;

    vertices[1].x = 0;
    vertices[1].y = 0;
    vertices[1].z = 0;
    slopes[1].dx =  pLeft->dy;
    slopes[1].dy = -pLeft->dx;
    slopes[1].dz =  pLeft->dy;
    slopes[1].k = 0;
    slopes[1].kz = 0;

    vertices[2].x = pLeft->xa;
    vertices[2].y = pLeft->ya;
    vertices[2].z = pLeft->za;

    adx = dx = pRight->xa - pLeft->xa;
    ady = dy = pRight->ya - pLeft->ya;
    adz = dz = pRight->za - pLeft->za;
    if (adx < 0)
        adx = -adx;
    if (ady < 0)
        ady = -ady;
    if (adz < 0)
        adz = -adz;
    scale = ady;
    if (adx > ady)
        scale = adx;
    if (adz > scale)
      scale = adz;
    slopes[2].dx = (dx * 65536) / scale;
    slopes[2].dy = (dy * 65536) / scale;
    slopes[2].dz = (dz * 65536) / scale;
    slopes[2].k = ((pLeft->xa + pRight->xa) * slopes[2].dy -
		   (pLeft->ya + pRight->ya) * slopes[2].dx) / 2.0;
    slopes[2].kz = ((pLeft->za + pRight->za) * slopes[2].dy -
		    (pLeft->ya + pRight->ya) * slopes[2].dz) / 2.0;
    edgecount = 3;

    y = miZPolyBuildPoly (vertices, slopes, edgecount, 
			 pLeft->x, pLeft->y, pLeft->z,
			 left, right, &nleft, &nright, &height);
    miZFillPolyHelper (zbs_ptr, pDrawable, pGC, pixel, spanData, y, height, left, right, nleft, nright);
}

static void
miZWideSegment (zbs_ptr, pDrawable, pGC, pixel, spanData,
		x1, y1, z1, x2, y2, z2, 
		projectLeft, projectRight, leftFace, rightFace)
    ddZBufferStruct *zbs_ptr;
    DrawablePtr	    pDrawable;
    GCPtr	    pGC;
    unsigned long   pixel;
    ZSpanDataPtr    spanData;
    register int    x1, y1, z1, x2, y2, z2;
    Bool	    projectLeft, projectRight;
    register ZLineFacePtr leftFace, rightFace; 
    /* NOTE: these are ARBITRARY NAMES-- left and right face are just the long
         sides of a wideline */ 
{
    double	l, L, r;
    double	xa, ya, za;
    double	projectXoff, projectYoff;
    double	k, kz;
    double	maxy;
    int		x, y, z;
    int		dx, dy, dz;
    int		finaly;
    ZPolyEdgePtr left, right;
    ZPolyEdgePtr top, bottom;
    int		lefty, righty, topy, bottomy;
    int		signdx, signdz;
    ZPolyEdgeRec lefts[2], rights[2];
    ZLineFacePtr tface;
    int		lw = pGC->lineWidth;

    /* Insure that lines always have a width of at least one pixel wide */
    if (lw == 0) lw = 1;
    /* draw top-to-bottom always */
    if (y2 < y1 || y2 == y1 && x2 < x1)
    {
	x = x1;
	x1 = x2;
	x2 = x;

	y = y1;
	y1 = y2;
	y2 = y;

	z = z1;
	z1 = z2;
	z2 = z;

	x = projectLeft;
	projectLeft = projectRight;
	projectRight = x;

	tface = leftFace;
	leftFace = rightFace;
	rightFace = tface;
    }

    dy = y2 - y1;
    signdx = 1;
    signdz = 1;
    dx = x2 - x1;
    dz = z2 - z1;
    if (dx < 0)
	signdx = -1;
    if (dz < 0)
	signdz = -1;

    leftFace->x = x1;   /* This information is not needed here.  It is used */
    leftFace->y = y1;   /*  by miZLineJoin */
    leftFace->z = z1;
    leftFace->dx = dx;
    leftFace->dy = dy;
    leftFace->dz = dz;

    rightFace->x = x2;
    rightFace->y = y2;
    rightFace->z = z2;
    rightFace->dx = -dx;
    rightFace->dy = -dy;
    rightFace->dz = -dz;
    
    leftFace->xa = 0;  /* In order to simplify things I moved this here */
    leftFace->ya = 0;  /* This is the default offset from the center of the */
    leftFace->za = 0;  /*  left and right face */
    rightFace->xa = 0;
    rightFace->ya = 0;
    rightFace->za = 0;

    /* If rectangle calls can be used, then do that.  Otherwise we have to
        do things the hard way */
    if (dy == 0)
    {   /* a horizontal line */
	rightFace->ya = (double) lw / 2.0;
	rightFace->k = -(double) (lw * dx) / 2.0;
	rightFace->kz = -(double) (lw * dz) / 2.0;
	leftFace->ya = -rightFace->ya;
	leftFace->k = rightFace->k;
	leftFace->kz = rightFace->kz;   /* Not sure what k, kz does.  This may
					   in fact be an incorrect value */
	x = x1;
	z = z1;
	if (projectLeft)
	    x -= (lw >> 1);
	y = y1 - (lw >> 1);
	dx = x2 - x;
	dz = z2 - z;
	if (projectRight)
	    dx += (lw + 1 >> 1);
	dy = lw;
	miZFillRectPolyHelper (zbs_ptr, pDrawable, pGC, pixel, spanData,
			      x, y, z, dx, dy, dz);
    }
    else if (dx == 0)
    {   /* a vertical line */
	leftFace->xa =  (double) lw / 2.0;
	leftFace->k = (double) (lw * dy) / 2.0;
	leftFace->kz = (double) (lw * dy) / 2.0;
	rightFace->xa = -leftFace->xa;
	rightFace->k = leftFace->k;
	rightFace->kz = leftFace->kz;   /* Not sure what k, kz does.  This may
					   in fact be an incorrect value */
	y = y1;
	if (projectLeft)
	    y -= lw >> 1;
	z = z1;
	x = x1 - (lw >> 1);
	dz = z2 - z;
	dy = y2 - y;
	if (projectRight)
	    dy += (lw + 1 >> 1);
	dx = lw;
	miZFillRectPolyHelper (zbs_ptr, pDrawable, pGC, pixel, spanData,
			      x, y, z, dx, dy, dz);
    }
    else
    {   /* a diagonal line */
    	l = ((double) lw) / 2.0;
    	L = hypot ((double) dx, (double) dy);

	if (dx < 0)
	{
	    right = &rights[1];
	    left = &lefts[0];
	    top = &rights[0];
	    bottom = &lefts[1];
	}
	else
	{
	    right = &rights[0];
	    left = &lefts[1];
	    top = &lefts[0];
	    bottom = &rights[1];
	}
	r = l / L;

	/* coord of upper bound at integral y */
	za = 0;                /* I am making the widelines have zero depth, 
				  regardless of orientation */
	ya = -r * dx;
	xa = r * dy;

	if (projectLeft | projectRight)
	{
	    projectXoff = -ya;
	    projectYoff = xa;
	}

    	/* xa * dy - ya * dx */
	k = l * L;
	kz = l * L;            /* Not sure what k, kz does.  This may
				  in fact be an incorrect value */

	leftFace->xa = xa;
	leftFace->ya = ya;
	leftFace->za = za;
	leftFace->k = k;
	leftFace->kz = kz;
	rightFace->xa = -x;
	rightFace->xa = -xa;
	rightFace->ya = -ya;
	rightFace->za = -za;
	rightFace->k = k;
	rightFace->kz = kz;

	if (projectLeft)
	    righty = miZPolyBuildEdge (xa - projectXoff, ya - projectYoff, za,
				      k, kz, dx, dy, dz, x1, y1, z1, 0, right);
	else
	    righty = miZPolyBuildEdge (xa, ya, za,
				      k, kz, dx, dy, dz, x1, y1, z1, 0, right);

	/* coord of lower bound at integral y */
	za = -za;
	ya = -ya;
	xa = -xa;

	/* xa * dy - ya * dx */
	k = - k;
	kz = - kz;

	if (projectLeft)
	    lefty = miZPolyBuildEdge (xa - projectXoff, ya - projectYoff, za,
				     k, kz, dx, dy, dz, x1, y1, z1, 1, left);
	else
	    lefty = miZPolyBuildEdge (xa, ya, za,
				     k, kz, dx, dy, dz, x1, y1, z1, 1, left);

	/* coord of top face at integral y */

	if (signdx > 0)
	{
	    za = -za;
	    ya = -ya;
	    xa = -xa;
	}
	if (projectLeft)
	{
	    double xap = xa - projectXoff;
	    double yap = ya - projectYoff;
	    topy = miZPolyBuildEdge (xap, yap, za, xap * dx + yap * dy, za * dz + yap * dy,
				    -dy, dx, dz, x1, y1, z1, dx > 0, top);
	}
	else
	    topy = miZPolyBuildEdge (xa, ya, za, 0.0, 0.0, -dy, dx, dz, x1, y1, z1, 
				    dx > 0, top);

	/* coord of bottom face at integral y */

	if (projectRight)
	{
	    double xap = xa + projectXoff;
	    double yap = ya + projectYoff;
	    bottomy = miZPolyBuildEdge (xap, yap, za, xap * dx + yap * dy, za * dz + yap * dy, 
				       -dy, dx, dz, x2, y2, z2, dx < 0, bottom);
	    maxy = -ya + projectYoff;
	}
	else
	{
	    bottomy = miZPolyBuildEdge (xa, ya, za,
				       0.0, 0.0, -dy, dx, dz, x2, y2, z2, 
				       dx < 0, bottom);
	    maxy = -ya;
	}

	finaly = ICEIL (maxy) + y2;

	if (dx < 0)
	{
	    left->height = bottomy - lefty;
	    right->height = finaly - righty;
	    top->height = righty - topy;
	}
	else
	{
	    right->height =  bottomy - righty;
	    left->height = finaly - lefty;
	    top->height = lefty - topy;
	}
	bottom->height = finaly - bottomy;
	miZFillPolyHelper (zbs_ptr, pDrawable, pGC, pixel, spanData, topy,
		     bottom->height + bottomy - topy, lefts, rights, 2, 2);
    }
}

ZSpanDataPtr
miZSetupSpanData (pGC, spanData, npt)
    register GCPtr pGC;
    ZSpanDataPtr spanData;
    int		npt;
{
    if (npt < 3 && pGC->capStyle != CapRound || miSpansEasyRopZ(pGC->alu))
	return (ZSpanDataPtr) NULL;
    if (pGC->lineStyle == LineDoubleDash)
	miInitSpanGroupZ (&spanData->bgGroup);
    miInitSpanGroupZ (&spanData->fgGroup);
    return spanData;
}

void
miZCleanupSpanData (zbs_ptr, pDrawable, pGC, spanData)
    ddZBufferStruct zbs_ptr;
    DrawablePtr	 pDrawable;
    GCPtr	 pGC;
    ZSpanDataPtr spanData;
{
    if (pGC->lineStyle == LineDoubleDash)
    {
	unsigned long	oldPixel, pixel;
	
	pixel = pGC->bgPixel;
	oldPixel = pGC->fgPixel;
    	if (pixel != oldPixel)
    	{
    	    DoChangeGC (pGC, GCForeground, (XID *)&pixel, FALSE);
    	    ValidateGC (pDrawable, pGC);
    	}
	miFillUniqueSpanGroupZ (zbs_ptr, pDrawable, pGC, &spanData->bgGroup);
	miFreeSpanGroupZ (&spanData->bgGroup);
    	if (pixel != oldPixel)
    	{
	    DoChangeGC (pGC, GCForeground, (XID *)&oldPixel, FALSE);
	    ValidateGC (pDrawable, pGC);
    	}
    }
    miFillUniqueSpanGroupZ (zbs_ptr, pDrawable, pGC, &spanData->fgGroup);
    miFreeSpanGroupZ (&spanData->fgGroup);
}

void
miZWideLine (zbs_ptr, pDrawable, pGC, mode, npt, pPts)
    ddZBufferStruct *zbs_ptr;
    DrawablePtr	pDrawable;
    register GCPtr pGC;
    int		mode;
    register int npt;
    register ddCoord3DS *pPts;
{
    int		    x1, y1, z1, x2, y2, z2;
    ZSpanDataRec    spanDataRec;
    ZSpanDataPtr    spanData;
    unsigned long   pixel;
    Bool	    projectLeft, projectRight;
    ZLineFaceRec    leftFace, rightFace, prevRightFace;
    ZLineFaceRec    firstFace;
    register int    first;
    Bool	    somethingDrawn = FALSE;
    Bool	    selfJoin;


    /* XXX backward compatibility */
    if (pGC->lineWidth == 0)
    {
	miZZeroLine (zbs_ptr, pDrawable, pGC, mode, npt, pPts);
	return;
    }
    spanData = miZSetupSpanData (pGC, &spanDataRec, npt);
    pixel = pGC->fgPixel;
    x2 = pPts->x;
    y2 = pPts->y;
    z2 = pPts->z;
    first = TRUE;
    selfJoin = FALSE;
    if (npt > 1)               /* check for self-joining polyline */
    {
    	if (mode == CoordModePrevious)
    	{
	    int nptTmp;
	    ddCoord3DS *pPtsTmp;
    
	    x1 = x2;
	    y1 = y2;
	    z1 = z2;
	    nptTmp = npt;
	    pPtsTmp = pPts + 1;
	    while (--nptTmp)
	    {
	    	x1 += pPtsTmp->x;
	    	y1 += pPtsTmp->y;
	    	z1 += pPtsTmp->z;
	    	++pPtsTmp;
	    }
	    if (x2 == x1 && y2 == y1 && z2 == z1)
	    	selfJoin = TRUE;
    	}
    	else if (x2 == pPts[npt-1].x && y2 == pPts[npt-1].y && z2 == pPts[npt-1].z)
    	{
	    selfJoin = TRUE;
    	}
    }
    projectLeft = pGC->capStyle == CapProjecting && !selfJoin;
    projectRight = FALSE;
    while (--npt)         /* loop through the points and draw each segment */
    {
	x1 = x2;
	y1 = y2;
	z1 = z2;
	++pPts;
	x2 = pPts->x;
	y2 = pPts->y;
	z2 = pPts->z;
	if (mode == CoordModePrevious)
	{
	    x2 += x1;
	    y2 += y1;
	    z2 += z1;
	}
	if (x1 != x2 || y1 != y2 || z1 != z2)
	{
	    somethingDrawn = TRUE;
	    if (npt == 1 && pGC->capStyle == CapProjecting && !selfJoin)
	    	projectRight = TRUE;
	    miZWideSegment (zbs_ptr, pDrawable, pGC, pixel, spanData, 
			    x1, y1, z1, x2, y2, z2, 
			    projectLeft, projectRight, &leftFace, &rightFace);
	    if (first)
	    {
	    	if (selfJoin)
		    firstFace = leftFace;
	    }
	    else
	    {            /* add a join if necessary in between segments*/
	    	miZLineJoin (zbs_ptr, pDrawable, pGC, pixel, spanData, &leftFace,
		            &prevRightFace);
	    }
	    prevRightFace = rightFace;
	    first = FALSE;
	    projectLeft = FALSE;
	}
	if (npt == 1 && somethingDrawn)
 	{
	    if (selfJoin)  /* add a join at the end if necessary */
		miZLineJoin (zbs_ptr, pDrawable, pGC, pixel, spanData, &firstFace,
			    &rightFace);
	}
    }
    /* Even if the line is a dot, we still need to draw something.  This 
        takes care of that case.  */
    if (!somethingDrawn)
    {
	projectLeft = pGC->capStyle == CapProjecting;
	miZWideSegment (zbs_ptr, pDrawable, pGC, pixel, spanData,
		       x2, y2, z2, x2, y2, z2, projectLeft, projectLeft,
		       &leftFace, &rightFace);
    }
    if (spanData)
	miZCleanupSpanData (zbs_ptr, pDrawable, pGC, spanData);
}

#define V_TOP	    0
#define V_RIGHT	    1
#define V_BOTTOM    2
#define V_LEFT	    3

static void
miZWideDashSegment (zbs_ptr, pDrawable, pGC, spanData, pDashOffset, pDashIndex,
	   x1, y1, z1, x2, y2, z2, projectLeft, projectRight, leftFace, rightFace)
    ddZBufferStruct zbs_ptr;
    DrawablePtr	    pDrawable;
    register GCPtr  pGC;
    int		    *pDashOffset, *pDashIndex;
    ZSpanDataPtr    spanData;
    int		    x1, y1, z1, x2, y2, z2;
    Bool	    projectLeft, projectRight;
    ZLineFacePtr    leftFace, rightFace;
{
    int		    dashIndex, dashRemain;
    unsigned char   *pDash;
    double	    L, l;
    double	    k, kz;
    ZPolyVertexRec  vertices[4];
    ZPolyVertexRec  saveRight, saveBottom;
    ZPolySlopeRec   slopes[4];
    ZPolyEdgeRec    left[2], right[2];
    ZLineFaceRec    lcapFace, rcapFace;
    int		    nleft, nright;
    int		    h;
    int		    y;
    int		    dz, dy, dx;
    unsigned long   pixel;
    double	    LRemain;
    double	    r;
    double	    rdx, rdy, rdz;
    double	    dashDx, dashDy, dashDz;
    double	    saveK;
    Bool	    first = TRUE;
    double	    lcenterx, lcentery, lcenterz, rcenterx, rcentery, rcenterz;
    unsigned long   fgPixel, bgPixel;
    
    dx = x2 - x1;
    dy = y2 - y1;
    dz = z2 - z1;
    dashIndex = *pDashIndex;
    pDash = pGC->dash;
    dashRemain = pDash[dashIndex] - *pDashOffset;
    fgPixel = pGC->fgPixel;
    bgPixel = pGC->bgPixel;
    if (pGC->fillStyle == FillOpaqueStippled ||
	pGC->fillStyle == FillTiled)
    {
	bgPixel = fgPixel;
    }

    l = ((double) pGC->lineWidth) / 2.0;
    if (dx == 0)
    {
	L = dy;
	rdx = 0;
	rdy = l;
	if (dy < 0)
	{
	    L = -dy;
	    rdy = -l;
	}
    }
    else if (dy == 0)
    {
	L = dx;
	rdx = l;
	rdy = 0;
	if (dx < 0)
	{
	    L = -dx;
	    rdx = -l;
	}
    }
    else
    {
	L = hypot ((double) dx, (double) dy);
	r = l / L;

	rdx = r * dx;
	rdy = r * dy;
    }
    k = l * L;
    kz = l * L;
    LRemain = L;
    /* All position comments are relative to a line with dx and dy > 0,
     * but the code does not depend on this */
    /* top */
    slopes[V_TOP].dx = dx;
    slopes[V_TOP].dy = dy;
    slopes[V_TOP].dz = dz;
    slopes[V_TOP].k = k;
    slopes[V_TOP].kz = kz;
    /* right */
    slopes[V_RIGHT].dx = -dy;
    slopes[V_RIGHT].dy = dx;
    slopes[V_RIGHT].dz = dz;
    slopes[V_RIGHT].k = 0;
    slopes[V_RIGHT].kz = 0;
    /* bottom */
    slopes[V_BOTTOM].dx = -dx;
    slopes[V_BOTTOM].dy = -dy;
    slopes[V_BOTTOM].dz = dz;
    slopes[V_BOTTOM].k = k;
    slopes[V_BOTTOM].kz = kz;
    /* left */
    slopes[V_LEFT].dx = dy;
    slopes[V_LEFT].dy = -dx;
    slopes[V_LEFT].dz = dz;
    slopes[V_LEFT].k = 0;
    slopes[V_LEFT].kz = 0;

    /* preload the start coordinates */
    vertices[V_RIGHT].x = vertices[V_TOP].x = rdy;
    vertices[V_RIGHT].y = vertices[V_TOP].y = -rdx;
/*    vertices[V_RIGHT].z = vertices[V_TOP].z = z;*/

    vertices[V_BOTTOM].x = vertices[V_LEFT].x = -rdy;
    vertices[V_BOTTOM].y = vertices[V_LEFT].y = rdx;
/*    vertices[V_BOTTOM].z = vertices[V_LEFT].z = rdx;*/

    if (projectLeft)
    {
	vertices[V_TOP].x -= rdx;
	vertices[V_TOP].y -= rdy;
/*	vertices[V_TOP].z -= rdy;*/

	vertices[V_LEFT].x -= rdx;
	vertices[V_LEFT].y -= rdy;
/*	vertices[V_LEFT].z -= rdy;*/

	slopes[V_LEFT].k = rdx * dx + rdy * dy;
/*	slopes[V_LEFT].kz = rdx * dx + rdy * dy;*/
    }

    lcenterx = x1;
    lcentery = y1;

    while (LRemain > dashRemain)
    {
	dashDx = (dashRemain * dx) / L;
	dashDy = (dashRemain * dy) / L;
	dashDz = (dashRemain * dz) / L;

	rcenterx = lcenterx + dashDx;
	rcentery = lcentery + dashDy;

	vertices[V_RIGHT].x += dashDx;
	vertices[V_RIGHT].y += dashDy;
	vertices[V_RIGHT].z += dashDz;

	vertices[V_BOTTOM].x += dashDx;
	vertices[V_BOTTOM].y += dashDy;
	vertices[V_BOTTOM].z += dashDz;

	slopes[V_RIGHT].k = vertices[V_RIGHT].x * dx + vertices[V_RIGHT].y * dy;
	slopes[V_RIGHT].kz = vertices[V_RIGHT].z * dz + vertices[V_RIGHT].y * dy;

	if (pGC->lineStyle == LineDoubleDash || !(dashIndex & 1))
	{
	    if (pGC->lineStyle == LineOnOffDash &&
 	        pGC->capStyle == CapProjecting)
	    {
		saveRight = vertices[V_RIGHT];
		saveBottom = vertices[V_BOTTOM];
		saveK = slopes[V_RIGHT].k;
		
		if (!first)
		{
		    vertices[V_TOP].x -= rdx;
		    vertices[V_TOP].y -= rdy;
/*		    vertices[V_TOP].z -= rdy;*/
    
		    vertices[V_LEFT].x -= rdx;
		    vertices[V_LEFT].y -= rdy;
/*		    vertices[V_LEFT].z -= rdz;*/

		    slopes[V_LEFT].k = vertices[V_LEFT].x *
				       slopes[V_LEFT].dy -
				       vertices[V_LEFT].y *
				       slopes[V_LEFT].dx;
		    slopes[V_LEFT].kz = vertices[V_LEFT].z *
				       slopes[V_LEFT].dy -
				       vertices[V_LEFT].y *
				       slopes[V_LEFT].dz;
		}
		
		vertices[V_RIGHT].x += rdx;
		vertices[V_RIGHT].y += rdy;
/*		vertices[V_RIGHT].z += rdy;*/

		vertices[V_BOTTOM].x += rdx;
		vertices[V_BOTTOM].y += rdy;
/*		vertices[V_BOTTOM].z += rdy;*/

		slopes[V_RIGHT].k = vertices[V_RIGHT].x *
				   slopes[V_RIGHT].dy -
				   vertices[V_RIGHT].y *
				   slopes[V_RIGHT].dx;
		slopes[V_RIGHT].kz = vertices[V_RIGHT].z *
				   slopes[V_RIGHT].dy -
				   vertices[V_RIGHT].y *
				   slopes[V_RIGHT].dz;
	    }
	    y = miZPolyBuildPoly (vertices, slopes, 4, x1, y1, z1,
			     	 left, right, &nleft, &nright, &h);
	    pixel = (dashIndex & 1) ? bgPixel : fgPixel;
	    miZFillPolyHelper (zbs_ptr, pDrawable, pGC, pixel, spanData, y, h, left, right, nleft, nright);

	    if (pGC->lineStyle == LineOnOffDash)
	    {
		switch (pGC->capStyle)
		{
		case CapProjecting:
		    vertices[V_BOTTOM] = saveBottom;
		    vertices[V_RIGHT] = saveRight;
		    slopes[V_RIGHT].k = saveK;
/*		    slopes[V_RIGHT].kz = saveK;*/
		    break;
	    	}
	    }
	}
	LRemain -= dashRemain;
	++dashIndex;
	if (dashIndex == pGC->numInDashList)
	    dashIndex = 0;
	dashRemain = pDash[dashIndex];

	lcenterx = rcenterx;
	lcentery = rcentery;

	vertices[V_TOP] = vertices[V_RIGHT];
	vertices[V_LEFT] = vertices[V_BOTTOM];
	slopes[V_LEFT].k = -slopes[V_RIGHT].k;
	first = FALSE;
    }

    if (pGC->lineStyle == LineDoubleDash || !(dashIndex & 1))
    {
    	vertices[V_TOP].x -= dx;
    	vertices[V_TOP].y -= dy;
    	vertices[V_TOP].z -= dz;

	vertices[V_LEFT].x -= dx;
	vertices[V_LEFT].y -= dy;
	vertices[V_LEFT].z -= dz;

	vertices[V_RIGHT].x = rdy;
	vertices[V_RIGHT].y = -rdx;
/*	vertices[V_RIGHT].z = -rdx;*/

	vertices[V_BOTTOM].x = -rdy;
	vertices[V_BOTTOM].y = rdx;
/*	vertices[V_BOTTOM].z = rdx;*/

	
	if (projectRight)
	{
	    vertices[V_RIGHT].x += rdx;
	    vertices[V_RIGHT].y += rdy;
/*	    vertices[V_RIGHT].z += rdy;*/
    
	    vertices[V_BOTTOM].x += rdx;
	    vertices[V_BOTTOM].y += rdy;
/*	    vertices[V_BOTTOM].z += rdy;*/
	    slopes[V_RIGHT].k = vertices[V_RIGHT].x *
				slopes[V_RIGHT].dy -
				vertices[V_RIGHT].y *
				slopes[V_RIGHT].dx;
	    slopes[V_RIGHT].kz = vertices[V_RIGHT].z *
				slopes[V_RIGHT].dy -
				vertices[V_RIGHT].y *
				slopes[V_RIGHT].dz;
	}
	else {
	    slopes[V_RIGHT].k = 0;
	    slopes[V_RIGHT].kz = 0;
	}

	if (!first && pGC->lineStyle == LineOnOffDash &&
	    pGC->capStyle == CapProjecting)
	{
	    vertices[V_TOP].x -= rdx;
	    vertices[V_TOP].y -= rdy;
/*	    vertices[V_TOP].z -= rdy;*/

	    vertices[V_LEFT].x -= rdx;
	    vertices[V_LEFT].y -= rdy;
/*	    vertices[V_LEFT].z -= rdy;*/
	    slopes[V_LEFT].k = vertices[V_LEFT].x *
			       slopes[V_LEFT].dy -
			       vertices[V_LEFT].y *
			       slopes[V_LEFT].dx;
	    slopes[V_LEFT].kz = vertices[V_LEFT].z *
			       slopes[V_LEFT].dy -
			       vertices[V_LEFT].y *
			       slopes[V_LEFT].dz;
	}
	else {
	    slopes[V_LEFT].k += dx * dx + dy * dy;
	    slopes[V_LEFT].kz += dz * dz + dy * dy;
	}


	y = miZPolyBuildPoly (vertices, slopes, 4, x2, y2, z2, 
			     left, right, &nleft, &nright, &h);

	pixel = (dashIndex & 1) ? pGC->bgPixel : pGC->fgPixel;
	miZFillPolyHelper (zbs_ptr, pDrawable, pGC, pixel, spanData, y, h, left, right, nleft, nright);
    }
    dashRemain = ((double) dashRemain) - LRemain;
    if (dashRemain == 0)
    {
	dashIndex++;
	if (dashIndex == pGC->numInDashList)
	    dashIndex = 0;
	dashRemain = pDash[dashIndex];
    }

    leftFace->x = x1;
    leftFace->y = y1;
    leftFace->z = z1;
    leftFace->dx = dx;
    leftFace->dy = dy;
    leftFace->dz = dz;
    leftFace->xa = rdy;
    leftFace->ya = -rdx;
/*    leftFace->za = -rdx;*/
    leftFace->k = k;
    leftFace->kz = kz;

    rightFace->x = x2;
    rightFace->y = y2;
    rightFace->z = z2;
    rightFace->dx = -dx;
    rightFace->dy = -dy;
    rightFace->dz = -dz;
    rightFace->xa = -rdy;
    rightFace->ya = rdx;
/*    rightFace->za = rdx;*/
    rightFace->k = k;
    rightFace->kz = kz;

    *pDashIndex = dashIndex;
    *pDashOffset = pDash[dashIndex] - dashRemain;
}

void
miZWideDash (zbs_ptr, pDrawable, pGC, mode, npt, pPts)
    ddZBufferStruct zbs_ptr;
    DrawablePtr	pDrawable;
    register GCPtr pGC;
    int		mode;
    register int npt;
    register ddCoord3DS *pPts;
{
    int		    x1, y1, z1, x2, y2, z2;
    unsigned long   pixel;
    Bool	    projectLeft, projectRight;
    ZLineFaceRec	    leftFace, rightFace, prevRightFace;
    ZLineFaceRec	    firstFace;
    int		    first;
    int		    dashIndex, dashOffset;
    register int    prevDashIndex;
    ZSpanDataRec	    spanDataRec;
    ZSpanDataPtr	    spanData;
    Bool	    somethingDrawn = FALSE;
    Bool	    selfJoin;
    Bool	    endIsFg, startIsFg, firstIsFg = FALSE, prevIsFg;

    /* XXX backward compatibility */
    if (pGC->lineWidth == 0)
    {
	miZZeroDashLine (zbs_ptr, pDrawable, pGC, mode, npt, pPts);
	return;
    }
    if (pGC->lineStyle == LineDoubleDash && 
	(pGC->fillStyle == FillOpaqueStippled || pGC->fillStyle == FillTiled))
    {
	miZWideLine (zbs_ptr, pDrawable, pGC, mode, npt, pPts);
	return;
    }
    if (npt == 0)
	return;
    spanData = miZSetupSpanData (pGC, &spanDataRec, npt);
    x2 = pPts->x;
    y2 = pPts->y;
    z2 = pPts->z;
    first = TRUE;
    selfJoin = FALSE;
    if (mode == CoordModePrevious)
    {
	int nptTmp;
	ddCoord3DS *pPtsTmp;

	x1 = x2;
	y1 = y2;
	z1 = z2;
	nptTmp = npt;
	pPtsTmp = pPts + 1;
	while (--nptTmp)
	{
	    x1 += pPtsTmp->x;
	    y1 += pPtsTmp->y;
	    z1 += pPtsTmp->z;
	    ++pPtsTmp;
	}
	if (x2 == x1 && y2 == y1 && z2 == z1)
	    selfJoin = TRUE;
    }
    else if (x2 == pPts[npt-1].x && y2 == pPts[npt-1].y && z2 == pPts[npt-1].z)
    {
	selfJoin = TRUE;
    }
    projectLeft = pGC->capStyle == CapProjecting && !selfJoin;
    projectRight = FALSE;
    dashIndex = 0;
    dashOffset = 0;
    miZStepDash ((int)pGC->dashOffset, &dashIndex,
	        pGC->dash, (int)pGC->numInDashList, &dashOffset);
    while (--npt)
    {
	x1 = x2;
	y1 = y2;
	z1 = z2;
	++pPts;
	x2 = pPts->x;
	y2 = pPts->y;
	z2 = pPts->z;
	if (mode == CoordModePrevious)
	{
	    x2 += x1;
	    y2 += y1;
	    z2 += z1;
	}
	if (x1 != x2 || y1 != y2 || z1 != z2)
	{
	    somethingDrawn = TRUE;
	    if (npt == 1 && pGC->capStyle == CapProjecting && 
		(!selfJoin || !firstIsFg))
		projectRight = TRUE;
	    prevDashIndex = dashIndex;
	    miZWideDashSegment (zbs_ptr, pDrawable, pGC, spanData, &dashOffset, &dashIndex,
				x1, y1, z1, x2, y2, z2,
				projectLeft, projectRight, &leftFace, &rightFace);
	    startIsFg = !(prevDashIndex & 1);
	    endIsFg = (dashIndex & 1) ^ (dashOffset != 0);
	    if (pGC->lineStyle == LineDoubleDash || startIsFg)
	    {
	    	pixel = startIsFg ? pGC->fgPixel : pGC->bgPixel;
	    	if (first || (pGC->lineStyle == LineOnOffDash && !prevIsFg))
	    	{
	    	    if (first && selfJoin)
		    {
		    	firstFace = leftFace;
			firstIsFg = startIsFg;
		    }
	    	}
	    	else
	    	{
	    	    miZLineJoin (zbs_ptr, pDrawable, pGC, pixel, spanData, &leftFace,
		            	&prevRightFace);
	    	}
	    }
	    prevRightFace = rightFace;
	    prevIsFg = endIsFg;
	    first = FALSE;
	    projectLeft = FALSE;
	}
	if (npt == 1 && somethingDrawn)
	{
	    if (pGC->lineStyle == LineDoubleDash || endIsFg)
	    {
		pixel = endIsFg ? pGC->fgPixel : pGC->bgPixel;
		if (selfJoin && (pGC->lineStyle == LineDoubleDash || firstIsFg))
		{
		    miZLineJoin (zbs_ptr, pDrawable, pGC, pixel, spanData, &firstFace,
				&rightFace);
		}
	    }
	    else
	    {
		/* glue a cap to the start of the line if
		 * we're OnOffDash and ended on odd dash
		 */
		if (selfJoin && firstIsFg)
		{
		    pixel = pGC->fgPixel;
		}
	    }
	}
    }
    /* handle crock where all points are coincident */
    if (!somethingDrawn && (pGC->lineStyle == LineDoubleDash || !(dashIndex & 1)))
    {
	/* not the same as endIsFg computation above */
	pixel = (dashIndex & 1) ? pGC->bgPixel : pGC->fgPixel;
	switch (pGC->capStyle) {
	case CapProjecting:
	    x1 = pGC->lineWidth;
	    miZFillRectPolyHelper (zbs_ptr, pDrawable, pGC, pixel, spanData,
				  x2 - (x1 >> 1), y2 - (x1 >> 1), z2, x1, x1);
	    break;
	}
    }
    if (spanData)
	miZCleanupSpanData (zbs_ptr, pDrawable, pGC, spanData);
}

/* these are stubs to allow old ddx ValidateGCs to work without change */

void
miZMiter()
{
}

void
miZNotMiter()
{
}
